#!/usr/bin/env python

import ujson
import asyncio
import aiofiles
import argparse
from elasticsearch.helpers import async_bulk
from copy import deepcopy
import os

from backup.base import BaseBackupWorker


class Loader(BaseBackupWorker):
    def __init__(self, mode: str, input_dir: str, connection_pool_size=5, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.input_dir = input_dir
        self.connection_pool_size = connection_pool_size
        self.mode = mode

    async def __aenter__(self):
        # create connection to elasticsearch
        await super().__aenter__()
        es_host, es_port = self.elastic_address.split(':')
        es_port = int(es_port)
        for _ in range(self.connection_pool_size - 1):
            self.es.transport.add_connection(dict(host=es_host, port=es_port))
        return self

    async def start(self):
        if self.mode == 'default':
            await self.create_index_with_meta()
        await self.upload_data()

    async def create_index_with_meta(self):
        async with aiofiles.open(os.path.join(self.input_dir, 'settings.json'), 'r') as sf,\
                aiofiles.open(os.path.join(self.input_dir, 'mappings.json'), 'r') as mf:
            settings, mappings = await asyncio.gather(sf.readline(), mf.readline())
    
        settings = ujson.loads(settings)['index']
        settings.pop('routing')
        settings.pop('provided_name')
        settings.pop('creation_date')
        settings.pop('uuid')
        settings.pop('version')
    
        mappings = ujson.loads(mappings)
        body = {"settings": settings, "mappings": mappings}

        res = await self.es.indices.create(self.index, body=body)
        self.logger.info('Index created sucessfully!' if res.get('acknowledged') else 'Index creation failed!')

    async def upload_data(self):
        inserts = []
        async with aiofiles.open(os.path.join(self.input_dir, 'data.json'), 'r') as f:
            actions, i = [], 0
            async for line in f:
                i += 1
                obj = ujson.loads(line)
                actions.append({"_index": self.index, "_id": str(obj['_id']), "_source": obj['_source']})
                if self.limit and i >= self.limit:
                    break
                if not i%self.chunk_size:
                    inserts.append(asyncio.create_task(async_bulk(self.es, deepcopy(actions))))
                    self.logger.info(f'{i} documents loaded')
                    actions.clear()
        if actions:
            inserts.append(asyncio.create_task(async_bulk(self.es, deepcopy(actions))))
        await asyncio.wait(inserts)
        self.logger.info('Data upload finished!')


async def main(args):
    async with Loader(
        elastic_address=args.elastic_address,
        index=args.index,
        input_dir=args.input_dir,
        chunk_size=args.chunk_size,
        limit=args.limit,
        connection_pool_size=args.connection_pool_size,
        mode=args.mode
    ) as loader:
        await loader.start()


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument('--elastic_address', type=str, required=True)
    parser.add_argument('--index', type=str, required=True)
    parser.add_argument('--input_dir', type=str, required=True)
    parser.add_argument('--limit', type=int, required=False, default=None)
    parser.add_argument('--chunk_size', type=int, required=False, default=500,\
        help='Insert `chunk_size` documents in a single bulk operation')
    parser.add_argument('--connection_pool_size', type=int, default=5)
    parser.add_argument('--mode', choices=['data', 'default'], required=False, default='default')
    args = parser.parse_args()

    asyncio.run(main(args))


if __name__ == '__main__':
    run()



import asyncio
import websockets
import argparse
from motor.motor_asyncio import AsyncIOMotorClient
import logging
import sys
import ujson
from pymongo import UpdateOne, DeleteOne, InsertOne
from copy import deepcopy
from bson import ObjectId

from mongo.elastic2mongo.base_streamer import BaseStreamer


class UpdateStreamer(BaseStreamer):

    def __init__(self, elastic_updates_uri: str, *args, **kwargs):
        """
        Args:
            elastic_updates_uri (str): Websocket address where changes are pushed to by means of `es-change-feed-plugin`
            More info at https://github.com/ForgeRock/es-change-feed-plugin
            Example: `ws://localhost:9400/ws/_changes`
        """
        super().__init__(*args, **kwargs)
        self.elastic_updates_uri = elastic_updates_uri
        self.updates = []
        self.mutex = asyncio.Lock()

    async def __aenter__(self):
        # create connection to mongodb
        self.mongo_client = AsyncIOMotorClient(self.mongo_address)
        self.mongo_col = self.mongo_client[self.mongo_db][self.mongo_collection]

        # add handlers for logger
        handler = logging.StreamHandler(sys.stdout)
        handler.setLevel(logging.INFO)
        handler.setFormatter(logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s'))
        self.logger.addHandler(handler)
        return self

    async def __aexit__(self, *exc_info):
        self.mongo_client.close()
        for handler in self.logger.handlers:
            handler.close()

    async def run(self):
        self.logger.info("Listening elasticsearch updates...")
        asyncio.create_task(self.periodic_push(period=1))
        async with websockets.connect(self.elastic_updates_uri) as websocket:
            async for message in websocket:
                asyncio.create_task(self.handle_change(message))

    async def handle_change(self, change: str):
        change = ujson.loads(change)
        if change['_index'] == self.elastic_index:
            operation = change['_operation']
            _id = ObjectId(change['_id']) if ObjectId.is_valid(change['_id']) else change['_id']
            if operation == 'CREATE':
                mongo_op = InsertOne(dict(**change['_source'], _id=_id))
            elif operation == 'INDEX':
                mongo_op = UpdateOne({'_id': _id}, {'$set': change['_source']}, upsert=True)
            elif operation == 'DELETE':
                mongo_op = DeleteOne({'_id': _id})
            else:
                return
            async with self.mutex:
                self.updates.append(mongo_op)
                if len(self.updates) == self.batch_size:
                    asyncio.create_task(self.write_to_db(deepcopy(self.updates)))
                    self.updates.clear()

    async def periodic_push(self, period=1):
        while True:
            await asyncio.sleep(period)
            async with self.mutex:
                if self.updates:
                    asyncio.create_task(self.write_to_db(deepcopy(self.updates)))
                    self.updates.clear()

    async def write_to_db(self, updates: list):
        res = await self.mongo_col.bulk_write(updates)
        self.logger.info(f"Bulk transaction result: {res.bulk_api_result}")


async def main(args):
    async with UpdateStreamer(
        elastic_updates_uri=args.elastic_updates_uri,
        mongo_address=args.mongo_address,
        mongo_db=args.mongo_db,
        mongo_collection=args.mongo_collection,
        elastic_address=args.elastic_address,
        elastic_index=args.elastic_index,
        batch_size=args.batch_size
    ) as streamer:
        await streamer.run()


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument('--mongo_address', type=str, required=True)
    parser.add_argument('--mongo_db', type=str, required=True)
    parser.add_argument('--mongo_collection', type=str, required=True)
    parser.add_argument('--elastic_address', type=str, required=True)
    parser.add_argument('--elastic_index', type=str, required=True)
    parser.add_argument('--batch_size', type=int, default=500)
    parser.add_argument('--elastic_updates_uri', type=str, required=True,\
        help='Websocket address where changes are pushed to by means of `es-change-feed-plugin`\
            More info at https://github.com/ForgeRock/es-change-feed-plugin')
    args = parser.parse_args()

    asyncio.run(main(args))


if __name__ == "__main__":
    run()


from mongo.abstract import AbstractStreamer
from elasticsearch import AsyncElasticsearch
from motor.motor_asyncio import AsyncIOMotorClient
import logging
import sys


class BaseStreamer(AbstractStreamer):

    def __init__(self, connection_pool_size: int, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.connection_pool_size = connection_pool_size

    async def __aenter__(self):
        # create connection to elasticsearch
        es_host, es_port = self.elastic_address.split(':')
        es_port = int(es_port)
        self.es = AsyncElasticsearch(self.elastic_address)
        for _ in range(self.connection_pool_size - 1):
            self.es.transport.add_connection(dict(host=es_host, port=es_port))

        # create connection to mongodb
        self.mongo_client = AsyncIOMotorClient(self.mongo_address)

        # add handlers for logger
        handler = logging.StreamHandler(sys.stdout)
        handler.setLevel(logging.INFO)
        handler.setFormatter(logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s'))
        self.logger.addHandler(handler)
        return self

    async def __aexit__(self, *exc_info):
        await self.es.close()
        self.mongo_client.close()
        for handler in self.logger.handlers:
            handler.close()



#!/usr/bin/env python

import asyncio
import argparse
import logging
import sys

from mongo.elastic2mongo.data_streamer import DataStreamer
from mongo.elastic2mongo.base_streamer import BaseStreamer


class Elastic2MongoStreamer(BaseStreamer):
    def __init__(self, *args, **kwargs):
        self.data_streamer = DataStreamer(*args, **kwargs)
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.setLevel(logging.INFO)

    async def __aenter__(self):
        # add handlers for logger
        handler = logging.StreamHandler(sys.stdout)
        handler.setLevel(logging.INFO)
        handler.setFormatter(logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s'))
        self.logger.addHandler(handler)
        return self

    async def __aexit__(self, *exc_info):
        for handler in self.logger.handlers:
            handler.close()

    async def run(self):
        async with self.data_streamer:
            self.logger.info('Streaming existing data...')
            await self.data_streamer.run()
            self.logger.info('Existing data streamed sucessfully')


async def main(args):
    async with Elastic2MongoStreamer(
        mongo_address=args.mongo_address,
        mongo_db=args.mongo_db,
        mongo_collection=args.mongo_collection,
        elastic_address=args.elastic_address,
        elastic_index=args.elastic_index,
        batch_size=args.batch_size
    ) as streamer:
        await streamer.run()


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument('--mongo_address', type=str, required=True)
    parser.add_argument('--mongo_db', type=str, required=True)
    parser.add_argument('--mongo_collection', type=str, required=True)
    parser.add_argument('--elastic_address', type=str, required=True)
    parser.add_argument('--elastic_index', type=str, required=True)
    parser.add_argument('--batch_size', type=int, default=500)
    args = parser.parse_args()

    asyncio.run(main(args))


if __name__ == "__main__":
    run()



#!/usr/bin/env python

import asyncio
from elasticsearch import AsyncElasticsearch
from elasticsearch.helpers import async_bulk
import argparse
import logging
import sys
from collections import deque
from motor.motor_asyncio import AsyncIOMotorClient
import traceback

from mongo.mongo2elastic.base_streamer import BaseStreamer


class UpdateStreamer(BaseStreamer):
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.actions = deque(maxlen=self.batch_size)
        self.mutex = asyncio.Lock()

    async def __aenter__(self):
        await super().__aenter__()
        self.mongo_col = self.mongo_client[self.mongo_db][self.mongo_collection]
        return self

    async def handle_change(self, change: dict):
        if change['operationType'] == 'delete':
            _id = change.get('documentKey', {}).get('_id')
            self.logger.info(f"New delete {_id}")
            res = await self.es.delete(index=self.elastic_index, id=_id)
        else:
            if change['operationType'] not in ['insert', 'replace', 'update']:
                return
            doc = change.get('fullDocument')
            try:
                self.logger.info(f"New {change['operationType']} {doc}")
                _id = doc.pop('_id')
                action = {
                    "_index": self.elastic_index,
                    "_id": str(_id),
                    "_source": doc
                }
                async with self.mutex:
                    self.actions.append(action)
                    if len(self.actions) == self.actions.maxlen:
                        res = await async_bulk(self.es, self.actions)
                        self.actions.clear()
                        self.logger.info(f"Bulk transaction result: {res}")
            except Exception as e:
                self.logger.error(f'{e}\n{traceback.format_exc()}')

    async def periodic_push(self, period=1):
        while True:
            await asyncio.sleep(period)
            if self.actions:
                async with self.mutex:
                    try:
                        res = await async_bulk(self.es, self.actions)
                        self.logger.info(f"Periodic bulk transaction result: {res}")
                    except Exception as e:
                        self.logger.error(f'{e}\n{traceback.format_exc()}')
                    self.actions.clear()

    async def run(self):
        async with self.mongo_col.watch(full_document='updateLookup') as stream:
            self.logger.info("Listening mongo updates...")
            asyncio.create_task(self.periodic_push(period=1))
            async for change in stream:
                asyncio.create_task(self.handle_change(change))


async def main(args):
    async with UpdateStreamer(
        mongo_address=args.mongo_address,
        mongo_db=args.mongo_db,
        mongo_collection=args.mongo_collection,
        elastic_address=args.elastic_address,
        elastic_index=args.elastic_index,
        batch_size=args.batch_size,
        connection_pool_size=args.connection_pool_size
    ) as streamer:
        await streamer.run()


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument('--mongo_address', type=str, required=True)
    parser.add_argument('--mongo_db', type=str, required=True)
    parser.add_argument('--mongo_collection', type=str, required=True)
    parser.add_argument('--elastic_address', type=str, required=True)
    parser.add_argument('--elastic_index', type=str, required=True)
    parser.add_argument('--batch_size', type=int, default=500)
    parser.add_argument('--connection_pool_size', type=int, default=5)
    args = parser.parse_args()

    asyncio.run(main(args))

if __name__ == '__main__':
    run()



#!/usr/bin/env python

import asyncio
import argparse
from bson import ObjectId

from mongo.elastic2mongo.base_streamer import BaseStreamer


class DataStreamer(BaseStreamer):
    
    async def run(self):
        tasks = []
        match_all = {
            "size": self.batch_size,
            "query": {
                "match_all": {}
            }
        }
        scroll_id = None

        while True:
            # We obtain scroll_id on the first run, then use it to get the next page of results and a new scroll_id
            scroll_response = await self.es.search(index=self.elastic_index, body=match_all, scroll='2s')\
                if not scroll_id else await self.es.scroll(scroll_id=scroll_id, scroll='2s')
            scroll_id = scroll_response['_scroll_id']

            documents = scroll_response['hits']['hits']
            if not documents:
                break
            documents = [dict(**d['_source'], _id=ObjectId(d['_id']) if ObjectId.is_valid(d['_id']) else d['_id'])\
                for d in documents]
            tasks.append(asyncio.create_task(self.insert_to_mongo(documents)))
        tasks.append(asyncio.create_task(self.es.clear_scroll(scroll_id=scroll_id)))
        await asyncio.wait(tasks)
        
    async def insert_to_mongo(self, documents: list):
        result = await self.mongo_client[self.mongo_db][self.mongo_collection].insert_many(documents)
        self.logger.info(f'Bulk insert result: {result}')

    
async def main(args):
    async with DataStreamer(
        mongo_address=args.mongo_address,
        mongo_db=args.mongo_db,
        mongo_collection=args.mongo_collection,
        elastic_address=args.elastic_address,
        elastic_index=args.elastic_index,
        batch_size=args.batch_size
    ) as streamer:
        await streamer.run()


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument('--mongo_address', type=str, required=True)
    parser.add_argument('--mongo_db', type=str, required=True)
    parser.add_argument('--mongo_collection', type=str, required=True)
    parser.add_argument('--elastic_address', type=str, required=True)
    parser.add_argument('--elastic_index', type=str, required=True)
    parser.add_argument('--batch_size', type=int, default=500)
    args = parser.parse_args()

    asyncio.run(main(args))


if __name__ == "__main__":
    run()



from abc import ABCMeta, abstractmethod
import logging


class AbstractStreamer(object, metaclass=ABCMeta):
    
    @abstractmethod
    def __init__(
        self,
        mongo_address: str,
        mongo_db: str,
        mongo_collection: str,
        elastic_address: str,
        elastic_index: str,
        batch_size: int = 500
    ):
        self.mongo_address = mongo_address
        self.mongo_db = mongo_db
        self.mongo_collection = mongo_collection
        self.elastic_address = elastic_address
        self.elastic_index = elastic_index
        self.batch_size = batch_size

        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.setLevel(logging.INFO)

    @abstractmethod
    async def run(self):
        pass

    @abstractmethod
    async def __aenter__(self):
        pass

    @abstractmethod
    async def __aexit__(self, *exc_info):
        pass



#!/usr/bin/env python

import asyncio
from elasticsearch import AsyncElasticsearch
from elasticsearch.helpers import async_bulk
import argparse
import logging
import sys
from collections import deque
from motor.motor_asyncio import AsyncIOMotorClient
from copy import deepcopy

from mongo.mongo2elastic.base_streamer import BaseStreamer


class DataStreamer(BaseStreamer):

    async def run(self):
        col = self.mongo_client[self.mongo_db][self.mongo_collection]

        cursor = col.find()
        actions = []
        tasks = []
        async for doc in cursor:
            _id = doc.pop('_id')
            actions.append({
                "_index": self.elastic_index,
                "_id": str(_id),
                "_source": doc
            })
            if len(actions) == self.batch_size:
                tasks.append(asyncio.create_task(self.stream_batch(deepcopy(actions))))
                actions.clear()
        if actions:
            tasks.append(asyncio.create_task(self.stream_batch(deepcopy(actions))))
        await asyncio.wait(tasks)

    async def stream_batch(self, batch: list):
        result = await async_bulk(self.es, batch)
        self.logger.info(f"Periodic bulk transaction result: {result}")
            

async def main(args):
    async with DataStreamer(
        mongo_address=args.mongo_address,
        mongo_db=args.mongo_db,
        mongo_collection=args.mongo_collection,
        elastic_address=args.elastic_address,
        elastic_index=args.elastic_index,
        batch_size=args.batch_size,
        connection_pool_size=args.connection_pool_size
    ) as streamer:
        await streamer.run()


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument('--mongo_address', type=str, required=True)
    parser.add_argument('--mongo_db', type=str, required=True)
    parser.add_argument('--mongo_collection', type=str, required=True)
    parser.add_argument('--elastic_address', type=str, required=True)
    parser.add_argument('--elastic_index', type=str, required=True)
    parser.add_argument('--batch_size', type=int, default=500)
    parser.add_argument('--connection_pool_size', type=int, default=5)
    args = parser.parse_args()

    asyncio.run(main(args))

if __name__ == '__main__':
    run()


from setuptools import setup, find_packages
from os import path


with open('requirements.txt') as f:
    required = f.read().splitlines()

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    version='1.0.1',
    name='elastic-tools',
    install_requires=required,
    packages=find_packages(),
    url='https://github.com/petricore-tech/elastictools.git',
    entry_points={
        'console_scripts': [
            'mongo2elastic=mongo.mongo2elastic:run',
            'elastic2mongo=mongo.elastic2mongo:run',
            'elasticdump=backup.dump:run',
            'elasticload=backup.load:run'
        ]
    },
    long_description=long_description,
    long_description_content_type='text/markdown'
)


import asyncio
from elasticsearch import AsyncElasticsearch
import aiofiles
from typing import Optional
import ujson
import os
import argparse
import logging
import sys
from abc import ABCMeta, abstractmethod


class BaseBackupWorker(object, metaclass=ABCMeta):

    def __init__(
        self, 
        elastic_address: str, 
        index: str, 
        chunk_size: int,
        limit: Optional[int] = None
    ):
        self.elastic_address = elastic_address
        self.index = index
        self.chunk_size = chunk_size
        self.limit = limit

        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.setLevel(logging.INFO)

    async def __aenter__(self):
        self.es = AsyncElasticsearch(hosts=self.elastic_address)
        self.logger.addHandler(logging.StreamHandler(sys.stdout))
        return self

    async def __aexit__(self, *exc_info):
        await self.es.close()
        [h.close() for h in self.logger.handlers]

    @abstractmethod
    async def start(self):
        pass
    


#!/usr/bin/env python

import asyncio
import aiofiles
import ujson
import os
import argparse

from backup.base import BaseBackupWorker


class Dumper(BaseBackupWorker):
    def __init__(self, output_dir, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.output_dir = output_dir
        
    async def dump_meta(self):
        meta = (await self.es.indices.get(self.index))[self.index]
        async with aiofiles.open(os.path.join(self.output_dir, 'mappings.json'), 'w') as mf, \
                aiofiles.open(os.path.join(self.output_dir, 'settings.json'), 'w') as sf:
            await asyncio.gather(
                mf.write(ujson.dumps(meta['mappings'])),
                sf.write(ujson.dumps(meta['settings']))
            )

    async def dump_data(self):
        count = self.limit or int((await self.es.cat.count(index=self.index)).strip().split()[-1])

        async with aiofiles.open(os.path.join(self.output_dir, 'data.json'), 'w') as f:
            tasks = []

            dumped_count = 0
            match_all = {
                "size": self.chunk_size,
                "query": {
                    "match_all": {}
                }
            }
            scroll_response = await self.es.search(index=self.index, body=match_all, scroll='2s')

            # write hits into file
            num_to_insert = min(self.limit - dumped_count, len(scroll_response['hits']['hits']))\
                if self.limit else len(scroll_response['hits']['hits'])
            documents = list(map(lambda x: f'{ujson.dumps(x)}\n',\
                scroll_response['hits']['hits'][:num_to_insert]))
            tasks.append(asyncio.create_task(f.writelines(documents)))
            
            # set scroll_id and update `dumped_count`
            scroll_id = scroll_response['_scroll_id']
            dumped_count += num_to_insert
            self.logger.info(f'{dumped_count} documents dumped')

            while dumped_count < count:
                scroll_response = await self.es.scroll(scroll_id=scroll_id, scroll='2s')
                # write hits into file
                num_to_insert = min(self.limit - dumped_count, len(scroll_response['hits']['hits']))\
                    if self.limit else len(scroll_response['hits']['hits'])

                documents = list(map(lambda x: f'{ujson.dumps(x)}\n',\
                    scroll_response['hits']['hits'][:num_to_insert]))
                tasks.append(asyncio.create_task(f.writelines(documents)))
                
                # set scroll_id and update `dumped_count`
                scroll_id = scroll_response['_scroll_id']
                dumped_count += num_to_insert

                self.logger.info(f'{dumped_count} documents dumped')
            tasks.append(asyncio.create_task(self.es.clear_scroll(scroll_id=scroll_id)))
            await asyncio.wait(tasks)
                    
    async def start(self):
        meta_task = asyncio.create_task(self.dump_meta())
        data_task = asyncio.create_task(self.dump_data())
        await asyncio.wait([meta_task, data_task])


async def main(args):
    async with Dumper(
        elastic_address=args.elastic_address, 
        index=args.index,
        output_dir=args.output_dir,
        chunk_size=args.chunk_size,
        limit=args.limit
    ) as dumper:
        await dumper.start()


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument('--elastic_address', type=str, required=True)
    parser.add_argument('--index', type=str, required=True)
    parser.add_argument('--output_dir', type=str, required=True)
    parser.add_argument('--chunk_size', type=int, required=False, default=500,\
        help='Get `chunk_size` documents from elasticsearch in a single request')
    parser.add_argument('--limit', type=int, required=False, default=None)
    args = parser.parse_args()

    asyncio.run(main(args))


if __name__ == "__main__":
    run()
